/*************************************************************************************/
/* program the Flash with data from C-Array						     */
/*************************************************************************************/

#define AT91M		1
#define AT91_PTRS	1
#include "at91xx.h"


#define LED_1	(1<<8)
#define LED_2	(1<<9)
#define LED_3	(1<<10)
#define LED_4	(1<<11)
#define LED_5	(1<<12)
#define LED_6	(1<<13)
#define LED_7	(1<<14)
#define LED_8	(1<<15)

#define ALL_LEDS LED_1|LED_2|LED_3|LED_4|LED_5|LED_6|LED_7|LED_8

#define LED1	(1<<26)
#define LED2	(1<<27)
#define LED3	(1<<28)
#define LED4	(1<<29)

#define TIMEOUT 200

/* invert pegel for LEDs */
/* #define INVERT */


#define WAIT 0x40000	/*set 0x1FFFF for about 25 MHz MCKI */

/*
error codes:
LED1:	DevCode || ManCode wrong
LED2:	error by eraseing
LED3:	error by programming BOOT
LED4:	error by programming APP
*/



/*************************************************************************************/
/* Function prototypes */

void show(int leds,int leds2, int count);
void ret_Code(int leds);
void set_leds(StructPIO *port,int leds);
void leds_on(StructPIO *port,int leds);
void leds_off(StructPIO *port,int leds);
int chip_erase();
void get_flash_id();
void sector_erase(long sector_addr);
int empty(long address, long size);
int program_word(long address,unsigned short word);
unsigned short read_word(long address);
void wait(void);

/*************************************************************************************/

StructPIO *at91_pio;
int DevCode;
int ManCode;

int main (void)
{
	long i,end,fl_adr=0;
	volatile unsigned short *dptr;
	int ok=1;

	at91_pio=PIOB_BASE;
	at91_pio->PIO_IDR  = 0xFFFFFFFF;	/* Disable PIO interrupts */
	at91_pio->PIO_PER  = ALL_LEDS;
	at91_pio->PIO_OER = ALL_LEDS;

	at91_pio=PIOA_BASE;
	at91_pio->PIO_IDR  = 0xFFFFFFFF;	/* Disable PIO interrupts */
	at91_pio->PIO_PER  = LED1 | LED2 | LED3 | LED4;
	at91_pio->PIO_OER = LED1 | LED2 | LED3 | LED4;

	set_leds(PIOB_BASE, 0);
	
	get_flash_id();	


	/* DeviceCode:
	     C0: AT49F16x4(8)		  C2: AT49F16x4T(8)  
	   00C0: AT49F16x4(16)		00C2: AT49F16x4T(16)  
	   
	   ManufacturerCode:
	     1F: (8) 	   001F: (16) 
	*/

	if (DevCode==0x00C0 && ManCode==0x001F)	
	{
		dptr= (unsigned short *)LOAD_FROM;
		end=*dptr<<16;
		end&=0xffff0000;
		dptr++;
		end+=*dptr;
		dptr++;		/* now --> points to data */

		/* check if flash free in prog-range */
		i=end;
		ok=empty((long)FLASHBASE, (long)i);

		if (ok!=0)
		{
			i=0;
			fl_adr=0xFFF;
			while(i<end)
			{
				i=fl_adr;
				sector_erase(fl_adr);
				set_leds(PIOB_BASE, (fl_adr&0xFF000)>>4);
				if (fl_adr<0x8000)
				{
					fl_adr+=0x1000;
					if (fl_adr>0x7fff)
					{
						fl_adr=0xBFFF;
					}
				}
				else
				{
					fl_adr+=0x4000;
				}
				/*
				i+=0xFFF;
				sector_erase((long)i);
				i+=1;
		
				erase so much sektors you need or erase the whole chip
					sector_erase(0xFFF);
					sector_erase(0x1FFF);
					sector_erase(0x2FFF);
					sector_erase(0x3FFF);
					sector_erase(0x4FFF);
					sector_erase(0x5FFF);
					sector_erase(0x6FFF);
					sector_erase(0x7FFF);
					sector_erase(0xBFFF);
					...
					 or erase full flash  --> ok=chip_erase(); 
				*/
			}
			/* check if flash free in prog-range */
			i=end;
			ok=empty((long)FLASHBASE, (long)i);
		}

		if (ok==0)
		{
			/* sign programming now FLASH */
			show(LED1|LED2, LED3|LED4, 5);
			for(i=0;i<10;i++)
				wait();

			for(i=0;i<end;i++)
			{
				ok=program_word( (long)i, (unsigned short)dptr[i]);
				if (ok!=0)
				{
					if (i<0xFFF)
						ret_Code(LED3);
					else
						ret_Code(LED4);
				}
			}

			ret_Code(LED1 | LED2 | LED3 | LED4);
		}
		else ret_Code(LED2);
	}
	else ret_Code(LED1);

	return(0);
}

/*************************************************************************************/

void show(int leds,int leds2, int count)
{
	while(count--)
	{
		leds_on(PIOA_BASE,leds);
		wait();
		leds_off(PIOA_BASE,leds);
		leds_on(PIOA_BASE,leds2);
		wait();
		leds_off(PIOA_BASE,leds2);
	}
}

/*************************************************************************************/

/* sign per LEDs */
void ret_Code(int leds)
{
	while(1)
	{
		leds_on(PIOA_BASE,leds);
		wait();
		leds_off(PIOA_BASE,leds);
		wait();
	}
}

/*************************************************************************************/

void set_leds(StructPIO *port,int leds)
{
	leds_off(port,~leds);
	leds_on(port,leds);
}

/*************************************************************************************/

void leds_on(StructPIO *port,int leds)
{
	at91_pio=port;
#ifdef INVERT
	at91_pio->PIO_CODR = leds;
#endif
#ifndef INVERT
	at91_pio->PIO_SODR = leds;
#endif
}

/*************************************************************************************/

void leds_off(StructPIO *port,int leds)
{
	at91_pio=port;
#ifdef INVERT
	at91_pio->PIO_SODR = leds;
#endif
#ifndef INVERT
	at91_pio->PIO_CODR = leds;
#endif
}

/*************************************************************************************/
void wait(void)
{
	int n;
	for(n=0;n<WAIT;n++) {n++;n--;}
}
/*************************************************************************************/
void __gccmain()
{
}

/*************************************************************************************/

void get_flash_id()
{
	volatile unsigned short *ptr= (unsigned short *)FLASHBASE;

	/* Enter Product Identification Mode */
	/* 3 Bus-Cycles */
	ptr[0x5555] = 0xaaaa;
	ptr[0x2aaa] = 0x5555;
	ptr[0x5555] = 0x9090;


	ManCode = (int)ptr[0];	/* A0=0 --> Manufacture Code */
	DevCode = (int)ptr[1];	/* A0=1 --> Device Code */

	/* Exit Product Identification Mode */
	/* 3 Bus-Cycles */
	ptr[0x5555] = 0xaaaa;
	ptr[0x2aaa] = 0x5555;
	ptr[0x5555] = 0xf0f0;
}

/*************************************************************************************/

int chip_erase()
{
	int i;
	volatile unsigned short *ptr= (unsigned short *)FLASHBASE;

	/* Chip Erase */
	/* 6 Bus-Cycles	*/
	ptr[0x5555] = 0x00aa;
	ptr[0x2aaa] = 0x0055;
	ptr[0x5555] = 0x0080;
	ptr[0x5555] = 0x00aa;
	ptr[0x2aaa] = 0x0055;
	ptr[0x5555] = 0x0010;
	
	i=0;
	while(i < 80)
	{
		
		leds_off(PIOA_BASE,LED1);
		wait();
		leds_on(PIOA_BASE,LED1);		

		leds_off(PIOA_BASE,LED2);
		wait();
		leds_on(PIOA_BASE,LED2);
		
		leds_off(PIOA_BASE,LED3);
		wait();
		leds_on(PIOA_BASE,LED3);
		
		leds_off(PIOA_BASE,LED4);
		wait();
		leds_on(PIOA_BASE,LED4);
		
		i++;
	}

	i = 1;
	/* chip needs max. 10s for erase */
	
	/* wait until Erase-process ready --> check Adr(0) with FFFFh */
	while(i < TIMEOUT)
	{
		if (ptr[0] != 0xffff)
		{
			i ++;
			wait();
		}
		else
		{
			i=0;
			break;
		}
	}

	return i;	/* return 0 --> successful */
}

/*************************************************************************************/

void sector_erase(long sector_addr)
{
	volatile unsigned short *ptr= (unsigned short *)FLASHBASE;

	/* Sector Erase */
	/* 6 Bus-Cycles	*/
	ptr[0x5555] = 0xaaaa;
	ptr[0x2aaa] = 0x5555;
	ptr[0x5555] = 0x8080;
	ptr[0x5555] = 0xaaaa;
	ptr[0x2aaa] = 0x5555;
	ptr[sector_addr] = 0x3030;	
	/* any byte/word-address within a sector can be used */
	/* NOTE: check datsheet if the sector has been protected */
	/* erase about 200ms*/
	wait();
	wait();
	wait();
	wait();
	wait();
}

/*************************************************************************************/
int empty(long address, long size)
{
	volatile unsigned short *ptr=(unsigned short *)address;
	while(size)
	{
		if (ptr[size] != 0xffff) return 1;
		size--;
	}
	return 0;
}

int program_word(long address,unsigned short word)
{
	int i;
	volatile unsigned short *ptr= (unsigned short *)FLASHBASE;

	/* Word Program */
	/* 4 Bus-Cycles	*/
	ptr[0x5555] = 0xaaaa;
	ptr[0x2aaa] = 0x5555;
	ptr[0x5555] = 0xa0a0;
	ptr[address] = word;

	for (i=0; i<200; i++)
	{
		i=i;
	}

	i=0;

	/* wait until Program-process ready --> check Adr(0) with FFFFh */
	while(i < TIMEOUT)
	{
		if (ptr[address] != word)
			i++;
		else
		{
			i=0;
			break;
		}
	}

	return i;	/* return 0 --> successful */
	
}

/*************************************************************************************/

unsigned short read_word(long address)
{
	volatile unsigned short *ptr= (unsigned short *)FLASHBASE;

	/* Read */
	/* 1 Bus-Cycles	*/

	return(ptr[address]);
}
